@@ -27,6 +27,8 @@
 #include "operations.h"
 #include "timeline.h"
 
+#include "cuda/cuda_kernels.h"
+
 namespace bluefog {
 namespace common {
 
@@ -503,29 +505,92 @@ void NCCLController::Broadcast(TensorTableEntry& entry) {
       });
 }
 
-void NCCLController::NeighborAllgather(TensorTableEntry& entry) {
-  // Note the order of recvcounts and displcments is by the oder of
-  // mpi_ctx_.neighbor_in_ranks_ because of MPI_Neighbor_allgather order is
-  // determined by then the order of the  values in sources and destinations is
-  // identical to the input that was used by the process with the same rank in
-  // comm_old in the creation call if the communicator was created with
-  // MPI_Dist_graph_create_adjacent.
-  int* recvcounts = new int[mpi_ctx_.neighbor_indgree_];
-  int* displcmnts = new int[mpi_ctx_.neighbor_indgree_];
-  if (!mpi_ctx_.IsTopoSetup()) {
-    throw std::runtime_error(
-        "Topology has not been set yet cannot run neighbor_allgather");
+void NCCLController::NeighborValueExchangeWithConstantElements(
+    const void* input_ptr, void* output_ptr, int num_elements, DataType dtype,
+    const std::vector<int>* dst_ranks, const std::vector<int>* src_ranks) {
+#if NCCL_MINOR > 6
+  int nsend = dst_ranks->size();
+  int nrecv = src_ranks->size();
+  ncclGroupStart();
+  for (int i = 0; i < nrecv; i++) {
+    int element_size = mpi_ctx_.GetMPITypeSize(dtype);  // Assume NCCL use same size as MPI
+    void* recvbuf =
+        (void*)(static_cast<char*>(output_ptr) + i * num_elements * element_size);
+    NCCLCHECK(ncclRecv(recvbuf, num_elements, GetNCCLDataType(dtype),
+                       src_ranks->at(i), nccl_ctx_.nccl_comm,
+                       nccl_ctx_.stream));
   }
-  Status status = mpi_ctx_.AllocateOutput(entry, recvcounts, Communicator::GRAPH);
-  mpi_ctx_.SetDisplacements(recvcounts, displcmnts, Communicator::GRAPH);
-  if (!CheckSameRecvSize(recvcounts, mpi_ctx_.neighbor_indgree_)) {
-    delete[] recvcounts;
-    delete[] displcmnts;
-    entry.callback(Status::PreconditionError(
-        "Neighbor_allgather/allreduce doesn't support varying lenght of "
-        "vector. Please make "
-        "sure the size of tensors is the same among all processes."));
-    return;
+  for (int i = 0; i < nsend; i++) {
+    NCCLCHECK(ncclSend(input_ptr, num_elements, GetNCCLDataType(dtype),
+                       dst_ranks->at(i), nccl_ctx_.nccl_comm, nccl_ctx_.stream));
+  }
+  ncclGroupEnd();
+#else
+  throw std::runtime_error("Try to use ncclSend and ncclRecv under NCCL < 2.7.");
+#endif
+}
+
+void NCCLController::NeighborValueExchangeWithVaryingElements(
+  const void* input_ptr, void* output_ptr,  const int sendcount,
+  const int* recvcounts, const int* displcmnts, DataType dtype,
+  const std::vector<int>* dst_ranks,
+  const std::vector<int>* src_ranks
+) {
+#if NCCL_MINOR > 6
+  int nsend = dst_ranks->size();
+  int nrecv = src_ranks->size();
+  ncclGroupStart();
+  for (int i = 0; i < nrecv; i++) {
+    int recv_count = recvcounts[i];
+    int target_disp = displcmnts[i];
+    int element_size = mpi_ctx_.GetMPITypeSize(dtype);  // Assume NCCL use same size as MPI
+    void* recvbuf =
+        (void*)(static_cast<char*>(output_ptr) + target_disp * element_size);
+    NCCLCHECK(ncclRecv(recvbuf, recv_count, GetNCCLDataType(dtype),
+                       src_ranks->at(i), nccl_ctx_.nccl_comm, nccl_ctx_.stream));
+  }
+  for (int i = 0; i < nsend; i++) {
+    NCCLCHECK(ncclSend(input_ptr, sendcount, GetNCCLDataType(dtype),
+                       dst_ranks->at(i), nccl_ctx_.nccl_comm, nccl_ctx_.stream));
+  }
+  ncclGroupEnd();
+#else
+  throw std::runtime_error("Try to use ncclSend and ncclRecv under NCCL < 2.7.");
+#endif
+}
+
+void NCCLController::NeighborAllgather(TensorTableEntry& entry) {
+  int* recvcounts;
+  int* displcmnts;
+  Status status;
+
+  if (!entry.dynamic_neighbors_enabled) {
+    if (!mpi_ctx_.IsTopoSetup()) {
+      throw std::runtime_error(
+          "Topology has not been set yet cannot run neighbor_allgather");
+    }
+    // Note the order of recvcounts and displcments is by the oder of
+    // mpi_ctx_.neighbor_in_ranks_ because of MPI_Neighbor_allgather order is
+    // determined by then the order of the  values in sources and destinations is
+    // identical to the input that was used by the process with the same rank in
+    // comm_old in the creation call if the communicator was created with
+    // MPI_Dist_graph_create_adjacent.
+    recvcounts = new int[mpi_ctx_.neighbor_indgree_];
+    displcmnts = new int[mpi_ctx_.neighbor_indgree_];
+    status = mpi_ctx_.AllocateOutput(entry, recvcounts, Communicator::GRAPH);
+  } else {
+    bool is_topo_check_fail = CheckNeighborSendRecvPatternForEntry(entry, mpi_ctx_, timeline_ptr_);
+    if (is_topo_check_fail) {
+      entry.callback(Status::InvalidArgument(
+          "Src and dst neighbor ranks do not match"));
+      return;
+    }
+
+    recvcounts = new int[entry.recv_neighbors->size()];
+    displcmnts = new int[entry.recv_neighbors->size()];
+    status = mpi_ctx_.AllocateOutput(entry, recvcounts, Communicator::DYNAMIC,
+                                     entry.send_neighbors.get(),
+                                     entry.recv_neighbors.get());
   }
   if (!status.ok()) {
     delete[] recvcounts;
@@ -534,6 +599,13 @@ void NCCLController::NeighborAllgather(TensorTableEntry& entry) {
     return;
   }
 
+  if (!entry.dynamic_neighbors_enabled) {
+    mpi_ctx_.SetDisplacements(recvcounts, displcmnts, Communicator::GRAPH);
+  } else {
+    mpi_ctx_.SetDisplacements(recvcounts, displcmnts, Communicator::DYNAMIC,
+                              /*source_neighbor_cnt=*/entry.recv_neighbors->size());
+  }
+
   const void* sendbuf = entry.tensor->data();
   int num_elements = entry.tensor->shape().num_elements();
   void* buffer_data = (void*)entry.output->data();
@@ -544,23 +616,17 @@ void NCCLController::NeighborAllgather(TensorTableEntry& entry) {
   timeline_ptr_->ActivityStart(entry.tensor_name, "COMM. (NCCL)");
   // Pitfall: neighbor_allgather do not include itself.
 #if NCCL_MINOR > 6
-  ncclGroupStart();
-  for (int i = 0; i < mpi_ctx_.neighbor_indgree_; i++) {
-    int recv_rank = mpi_ctx_.neighbor_in_ranks_[i];
-    int recv_count = recvcounts[i];
-    int target_disp = displcmnts[i];
-    int element_size = mpi_ctx_.GetMPITypeSize(
-        entry.tensor->dtype());  // Assume NCCL use same size as MPI
-    void* recvbuf =
-        (void*)(static_cast<char*>(buffer_data) + target_disp * element_size);
-    NCCLCHECK(ncclRecv(recvbuf, recv_count, GetNCCLDataType(entry.tensor),
-                       recv_rank, nccl_ctx_.nccl_comm, nccl_ctx_.stream));
+  if (!entry.dynamic_neighbors_enabled) {
+    NeighborValueExchangeWithVaryingElements(
+        sendbuf, buffer_data, num_elements, recvcounts, displcmnts,
+        entry.tensor->dtype(), &mpi_ctx_.neighbor_out_ranks_,
+        &mpi_ctx_.neighbor_in_ranks_);
+  } else {
+    NeighborValueExchangeWithVaryingElements(
+        sendbuf, buffer_data, num_elements, recvcounts, displcmnts,
+        entry.tensor->dtype(), entry.send_neighbors.get(),
+        entry.recv_neighbors.get());
   }
-  for (int send_rank : mpi_ctx_.neighbor_out_ranks_) {
-    NCCLCHECK(ncclSend(sendbuf, num_elements, GetNCCLDataType(entry.tensor),
-                       send_rank, nccl_ctx_.nccl_comm, nccl_ctx_.stream));
-  }
-  ncclGroupEnd();
 
   auto tid = std::this_thread::get_id();
   nccl_ctx_.finalizer_thread_pool.execute([this, entry, tid]() mutable {
@@ -661,9 +727,7 @@ void NCCLController::NeighborAllreduce(TensorTableEntry& entry) {
   // If only partial sending is enabled, the following code block checks whether
   // the sending and recieving neighbors match each other when enable_topo_check
   // is set to be True.
-  bool is_topo_check_fail = CheckNeighborSendRecvPattern(
-      mpi_ctx_.size_, entry, timeline_ptr_,
-      mpi_ctx_.GetMPICommunicator(Communicator::GLOBAL));
+  bool is_topo_check_fail = CheckNeighborSendRecvPatternForEntry(entry, mpi_ctx_, timeline_ptr_);
   if (is_topo_check_fail) {
     entry.callback(Status::InvalidArgument(
         "Send and recv neighbors dont' match in neighbor "
@@ -671,10 +735,13 @@ void NCCLController::NeighborAllreduce(TensorTableEntry& entry) {
     return; 
   }
 
+  std::vector<std::shared_ptr<common::Tensor>> weighted_tensors;
+  // Ensure the lifecycle of the weighted tensors are alive after communication.
+
 #if NCCL_MINOR > 6
-  if (!entry.is_hierarchical) {
+  if (!entry.is_hierarchical) {  // neighbor allreduce without hierarchy
     ncclGroupStart();
-    if (!entry.dynamic_neighbors_enabled) {
+    if (!entry.dynamic_neighbors_enabled) {  // static topology
       for (int i = 0; i < mpi_ctx_.neighbor_indgree_; i++) {
         int recv_rank = mpi_ctx_.neighbor_in_ranks_[i];
         void* recvbuf = (void*)(static_cast<const char*>(entry.output->data()) +
@@ -686,7 +753,19 @@ void NCCLController::NeighborAllreduce(TensorTableEntry& entry) {
         NCCLCHECK(ncclSend(sendbuf, num_elements, GetNCCLDataType(entry.tensor),
                           send_rank, nccl_ctx_.nccl_comm, nccl_ctx_.stream));
       }
-    } else {
+    } else {  // dynamic topology
+      if(entry.dst_weighting_enabled) {
+        for (size_t i = 0; i < entry.send_neighbors->size(); ++i) {
+          auto weighted_tensor_ptr = entry.tensor->data_weight(entry.send_weights->at(i));
+          weighted_tensors.push_back(std::move(weighted_tensor_ptr));
+        }
+      }
+      // TODO(ybc) #83 Better design pattern for data_weight synchronization
+      // This ready event makes sure the data_weight computation is done before communication, as
+      // Pytorch CUDA stream is not synchronized with our CUDA stream, and it does nothing when
+      // it is running on CPU.
+      std::shared_ptr<common::ReadyEvent> ready_event =
+          entry.context->RecordReadyEvent(entry.device);
       for (size_t i = 0; i < entry.recv_neighbors->size(); ++i) {
         int recv_rank = entry.recv_neighbors->at(i);
         void* recvbuf = (void*)(static_cast<const char*>(entry.output->data()) +
@@ -694,19 +773,32 @@ void NCCLController::NeighborAllreduce(TensorTableEntry& entry) {
         NCCLCHECK(ncclRecv(recvbuf, num_elements, GetNCCLDataType(entry.tensor),
                           recv_rank, nccl_ctx_.nccl_comm, nccl_ctx_.stream));
       }
-      for (int send_rank : *entry.send_neighbors) {
-        NCCLCHECK(ncclSend(sendbuf, num_elements, GetNCCLDataType(entry.tensor),
-                          send_rank, nccl_ctx_.nccl_comm, nccl_ctx_.stream));
+      if(entry.dst_weighting_enabled) {
+        while ((ready_event != nullptr) && !ready_event->Ready()) {
+          std::this_thread::sleep_for(std::chrono::nanoseconds(100));
+        }
+      }
+      for (size_t i = 0; i < entry.send_neighbors->size(); ++i) {
+        const void* buffer_send = (entry.dst_weighting_enabled)
+                                      ? weighted_tensors[i]->data()
+                                      : sendbuf;
+        NCCLCHECK(ncclSend(buffer_send, num_elements, GetNCCLDataType(entry.tensor),
+                           entry.send_neighbors->at(i), nccl_ctx_.nccl_comm, nccl_ctx_.stream));
       }
     }
     ncclGroupEnd();
-  } else {
+  } else {  // hierarchical neighbor allreduce
     if (entry.send_neighbors->empty()) {
       throw std::runtime_error(
           "Under hierarchical neighbor_allreduce, argument "
           "send_machine_neighbors should "
           "not be empty.");
     }
+    if (entry.dst_weighting_enabled) {
+      throw std::runtime_error(
+          "Under hierarchical neighbor_allreduce, argument "
+          "dst_weight should not be enabled for now.");
+    }
     // 1. In-place allreduce for all local ranks. Note it is sum, so we need to
     // divided by local size at call back stage.
     NCCLCHECK(ncclAllReduce(sendbuf, (void*)sendbuf, num_elements,
@@ -746,7 +838,7 @@ void NCCLController::NeighborAllreduce(TensorTableEntry& entry) {
 
   auto tid = std::this_thread::get_id();
   nccl_ctx_.finalizer_thread_pool.execute(
-      [this, entry, event_queue, tid]() mutable {
+      [this, entry, event_queue, tid, weighted_tensors]() mutable {
         with_device device_guard(entry.device);
         WaitForEvents(event_queue, {entry}, this->timeline_ptr_, tid);
 
@@ -907,9 +999,8 @@ void NCCLController::NeighborAllreduce(std::vector<TensorTableEntry>& entries) {
   // If only partial sending is enabled, the following code block checks whether
   // the sending and recieving neighbors match each other when enable_topo_check
   // is set to be True.
-  bool is_topo_check_fail = CheckNeighborSendRecvPattern(
-      mpi_ctx_.size_, first_entry, timeline_ptr_,
-      mpi_ctx_.GetMPICommunicator(Communicator::GLOBAL));
+  bool is_topo_check_fail = CheckNeighborSendRecvPatternForEntry(first_entry, mpi_ctx_,
+                                                                 timeline_ptr_);
   if (is_topo_check_fail) {
     for (auto& entry : entries) {
       entry.callback(Status::InvalidArgument(
@@ -927,6 +1018,12 @@ void NCCLController::NeighborAllreduce(std::vector<TensorTableEntry>& entries) {
     RecordEvent(event_queue, "MEM_CPY_IN");
   }
 
+  const void* weighted_fused_input_data =
+      (first_entry.dst_weighting_enabled)
+          ? GenerateWeightedFusedInputData(fused_input_data, first_entry,
+                                           num_elements, element_size)
+          : nullptr;
+
   // Unlike allreduce, the storage for neighbor_allreduce in fusion buffer
   // is like [t_1, t_2 | t_1_n1, t_2_n1, t_1_n2, t_2_n2].
   // Here t_1 and t_2  means self tensor 1 and 2 and _n1 and _n2 means the
@@ -934,9 +1031,9 @@ void NCCLController::NeighborAllreduce(std::vector<TensorTableEntry>& entries) {
   // Hence, we need to offset the buffer data to location for neighbors.
   buffer_data = (uint8_t*)buffer_data + num_elements * element_size;
 
-  if (!first_entry.is_hierarchical) {
+  if (!first_entry.is_hierarchical) {  // neighbor allreduce without hierarchy
     ncclGroupStart();
-    if (!first_entry.dynamic_neighbors_enabled) {
+    if (!first_entry.dynamic_neighbors_enabled) {  // static topology
       for (int i = 0; i < mpi_ctx_.neighbor_indgree_; i++) {
         int recv_rank = mpi_ctx_.neighbor_in_ranks_[i];
         void* recvbuf =
@@ -950,7 +1047,7 @@ void NCCLController::NeighborAllreduce(std::vector<TensorTableEntry>& entries) {
                            GetNCCLDataType(first_entry.tensor), send_rank,
                            nccl_ctx_.nccl_comm, nccl_ctx_.stream));
       }
-    } else {
+    } else {  // dynamic topology
       for (size_t i = 0; i < first_entry.recv_neighbors->size(); ++i) {
         int recv_rank = first_entry.recv_neighbors->at(i);
         void* recvbuf =
@@ -959,14 +1056,18 @@ void NCCLController::NeighborAllreduce(std::vector<TensorTableEntry>& entries) {
                            GetNCCLDataType(first_entry.tensor), recv_rank,
                            nccl_ctx_.nccl_comm, nccl_ctx_.stream));
       }
-      for (int send_rank : *first_entry.send_neighbors) {
-        NCCLCHECK(ncclSend(fused_input_data, num_elements,
-                           GetNCCLDataType(first_entry.tensor), send_rank,
+      for (size_t i = 0; i < first_entry.send_neighbors->size(); ++i) {
+        const void* sendbuf =
+            (first_entry.dst_weighting_enabled)
+                ? (void*)((uint8_t*)weighted_fused_input_data + num_elements * i * element_size)
+                : fused_input_data;
+        NCCLCHECK(ncclSend(sendbuf, num_elements, GetNCCLDataType(first_entry.tensor),
+                           first_entry.send_neighbors->at(i),
                            nccl_ctx_.nccl_comm, nccl_ctx_.stream));
       }
     }
     ncclGroupEnd();
-  } else {
+  } else {  // hierarchical neighbor allreduce
     if (first_entry.send_neighbors->empty()) {
       throw std::runtime_error(
           "Under hierarchical neighbor_allreduce, argument "
@@ -978,6 +1079,11 @@ void NCCLController::NeighborAllreduce(std::vector<TensorTableEntry>& entries) {
         "neighbor_allreduce instead of hierarchical_neighbor_allreduce."
       );
     }
+    if (first_entry.dst_weighting_enabled) {
+      throw std::runtime_error(
+          "Under hierarchical neighbor_allreduce, argument "
+          "dst_weight should not be enabled for now.");
+    }
 
     // 1. In-place allreduce for all local ranks. Note it is sum, so we need to
     // divided by local size at call back stage.
@@ -1037,7 +1143,7 @@ void NCCLController::NeighborAllreduce(std::vector<TensorTableEntry>& entries) {
 
   auto tid = std::this_thread::get_id();
   nccl_ctx_.finalizer_thread_pool.execute(
-      [this, entries, event_queue, tid, buffer_data]() mutable {
+      [this, entries, event_queue, tid, buffer_data, weighted_fused_input_data]() mutable {
         auto& first_entry = entries[0];
         with_device device_guard(first_entry.device);
         WaitForEvents(event_queue, entries, this->timeline_ptr_, tid);
@@ -1787,6 +1893,52 @@ void NCCLController::MemcpyInFusionBuffer(
   buffer_len = (size_t)offset;
 }
 
+void NCCLController::MemcpyInWeightFusionBuffer(
+  void*& weight_buffer_data, size_t num_dst,
+  const void* buffer_data, int64_t num_elements, int element_size,
+  std::shared_ptr<OpContext> context, int device) {
+  // Access the fusion buffer.
+  FusionBufferManager* buffer_manager;
+  auto fusion_status = GetBluefogFusionBuffer(buffer_manager);
+  if (!fusion_status.ok()){
+    throw std::runtime_error(fusion_status.reason());
+  }
+  std::shared_ptr<PersistentBuffer> buffer =
+      buffer_manager->GetWeightBuffer(device);
+  weight_buffer_data = const_cast<void*>(buffer->AccessData(context));
+  size_t data_size = num_elements * element_size;
+
+  int64_t offset = 0;
+  for (size_t i = 0; i < num_dst; ++i) {
+    void* weight_buffer_data_at_offset = (uint8_t*)weight_buffer_data + offset;
+    CUDACHECK(cudaMemcpyAsync(weight_buffer_data_at_offset, buffer_data, data_size,
+                              cudaMemcpyDeviceToDevice, nccl_ctx_.stream));
+    offset += data_size;
+  }
+}
+
+const void* NCCLController::GenerateWeightedFusedInputData(const void* fused_input_data,
+                                                           const TensorTableEntry& entry,
+                                                           int64_t num_elements, int element_size) {
+  // Given a fused_input_data like [t_1, t_2], the storage for neighbor_allreduce in
+  // weighted fused input data is like [t_1_w1, t_2_w1 | t_1_w2, t_2_w2 | t_1_w3, t_2_w3].
+  // Here t_1 and t_2  means self tensor 1 and 2 and _w1, _w2, and _w3 means the
+  // destination weights to destination 1, 2, and 3.
+  void* weight_buffer_data = nullptr;
+  MemcpyInWeightFusionBuffer(weight_buffer_data, entry.send_neighbors->size(),
+                             fused_input_data, num_elements, element_size,
+                             entry.context, entry.device);
+  int64_t offset = 0;
+  for (size_t i = 0; i < entry.send_neighbors->size(); ++i) {
+    double dst_weight = entry.send_weights->at(i);
+    void* weight_buffer_data_offset = (uint8_t*)weight_buffer_data + offset;
+    ScaleBufferCudaImpl(dst_weight, weight_buffer_data_offset, num_elements,
+                        entry.tensor->dtype(), nccl_ctx_.stream);
+    offset += num_elements * element_size;
+  }
+  return weight_buffer_data;
+}
+
 void NCCLController::MemcpyOutFusionBuffer(
     const void* buffer_data, std::vector<TensorTableEntry>& entries) {
   int64_t offset = 0;
