@@ -176,6 +176,13 @@ class NCCLController {
   ncclResult_t ncclRecvByBcast(void* sendbuf, const int count,
                                ncclDataType_t data_type, int peer_rank);
 #endif
+  void NeighborValueExchangeWithConstantElements(
+      const void* input_ptr, void* output_ptr, int num_elements, DataType dtype,
+      const std::vector<int>* dst_ranks, const std::vector<int>* src_ranks);
+  void NeighborValueExchangeWithVaryingElements(
+      const void* input_ptr, void* output_ptr, const int sendcount,
+      const int* recvcounts, const int* displcmnts, DataType dtype,
+      const std::vector<int>* dst_ranks, const std::vector<int>* src_ranks);
 
   void MemcpyInFusionBuffer(const std::vector<TensorTableEntry>& entries,
                             void*& buffer_data, size_t& buffer_len);
@@ -188,6 +195,14 @@ class NCCLController {
                                          const int num_recv_neighbors,
                                          const int64_t fused_data_size);
 
+  void MemcpyInWeightFusionBuffer(void*& weight_buffer_data, size_t num_dst,
+                                  const void* buffer_data, int64_t num_elements, int element_size,
+                                  std::shared_ptr<OpContext> context, int device);
+
+  const void* GenerateWeightedFusedInputData(const void* fused_input_data,
+                                             const TensorTableEntry& entry,
+                                             int64_t num_elements, int element_size);
+
   void MemcpyOutFusionBufferForInputs(const void* fused_input_data,
                                       std::vector<TensorTableEntry>& entries);
 
