@@ -63,6 +63,7 @@ NCCLContext nccl_context;
 // If set, win_ops will execute the same ops on associated p as well.
 static bool global_with_associated_p_state = false;
 static bool global_skip_negotiate_stage = false;
+static bool global_background_thread_suspend = false;
 
 const auto SUSPEND_BACKGROUND_WAITTING_DURATION = std::chrono::microseconds(10);
 
@@ -431,6 +432,20 @@ void CheckForStalledTensors(BluefogGlobalState& state) {
   }
 }
 
+template<typename T>
+bool IsSameList (std::shared_ptr<std::vector<T>> n1, std::shared_ptr<std::vector<T>> n2) {
+  if (n1 == nullptr && n2 == nullptr) return true;
+  if (n1 == nullptr || n2 == nullptr) return false;
+  if (n1->size() != n2->size()) return false;
+  // The order matters as well.
+  for (size_t i = 0; i < n1->size(); i++) {
+    if (n1->at(i) != n2->at(i)) {
+      return false;
+    }
+  }
+  return true;
+}
+
 }  // namespace
 
 bool RunLoopOnce(BluefogGlobalState& state);
@@ -467,6 +482,8 @@ void BackgroundThreadLoop(BluefogGlobalState& state) {
   if (bluefog_fusion_threshold != nullptr) {
     state.tensor_fusion_threshold =
         std::strtol(bluefog_fusion_threshold, nullptr, 10);
+    state.tensor_fusion_threshold_for_dst_weight =
+        state.tensor_fusion_threshold;
   }
 
   // Initialize the tensor count table. No tensors are available yet.
@@ -764,7 +781,19 @@ void PerformOperationWithFusion(std::vector<TensorTableEntry>& entries) {
       first_entry.context,
       [&]() { timeline.ActivityStartAll(entries, "INIT_FUSION_BUFFER"); },
       [&]() { timeline.ActivityEndAll(entries); });
-  if (!status.ok()) {
+  
+  // As the dst_weight requires extra memory to scale the tensor for each destination, therefore,
+  // extra memory is required.
+  Status status_dst_weight = Status::OK();
+  if (first_entry.dst_weighting_enabled) {
+    status_dst_weight = bluefog_global.fusion_buffer.InitializeWeightBuffer(
+      bluefog_global.tensor_fusion_threshold_for_dst_weight, mpi_context.size_,
+      first_entry.device, first_entry.context,
+      [&]() { timeline.ActivityStartAll(entries, "INIT_WEIGHT_FUSION_BUFFER"); },
+      [&]() { timeline.ActivityEndAll(entries); });
+  }
+
+  if (!status.ok() || !status_dst_weight.ok()) {
     for (auto& e : entries) {
       e.callback(status);
     }
@@ -824,7 +853,9 @@ void PerformOperationWithFusion(std::vector<TensorTableEntry>& entries) {
 void NegotiateOfRequestOfMaster(BluefogGlobalState& state,
                                 std::deque<Request>& message_queue_buffer,
                                 bool& should_change_topo,
-                                bool& should_shut_down) {
+                                bool& should_shut_down) {  
+  state.unfinished_enqueued_entries.fetch_add(message_queue_buffer.size());
+
   std::vector<std::string> ready_to_reduce;
   RequestList message_list;
   message_list.set_shutdown(should_shut_down);
@@ -947,20 +978,6 @@ void NegotiateOfRequestOfMaster(BluefogGlobalState& state,
       // Attempt to add more responses to this fused response.
       const TensorTableEntry& entry =
           state.tensor_queue.GetTensorEntry(response.tensor_names()[0]);
-      auto IsSameNeighborList =
-          [](std::shared_ptr<std::vector<int>> n1,
-             std::shared_ptr<std::vector<int>> n2) -> bool {
-        if (n1 == nullptr && n2 == nullptr) return true;
-        if (n1 == nullptr || n2 == nullptr) return false;
-        if (n1->size() != n2->size()) return false;
-        // The order matters as well.
-        for (int i = 0; i < n1->size(); i++) {
-          if (n1->at(i) != n2->at(i)) {
-            return false;
-          }
-        }
-        return true;
-      };
       // Recall that send_neighbors is empty or not determines we use partial
       // neighbor allreduce or not.
       int num_recv_neighbors = !entry.dynamic_neighbors_enabled
@@ -983,11 +1000,11 @@ void NegotiateOfRequestOfMaster(BluefogGlobalState& state,
             response.devices() == new_response.devices() &&
             entry.tensor->dtype() == new_entry.tensor->dtype() &&
             entry.dynamic_neighbors_enabled == new_entry.dynamic_neighbors_enabled &&
+            entry.dst_weighting_enabled == new_entry.dst_weighting_enabled &&
             entry.is_hierarchical == new_entry.is_hierarchical &&
-            IsSameNeighborList(entry.send_neighbors,
-                               new_entry.send_neighbors) &&
-            IsSameNeighborList(entry.recv_neighbors,
-                               new_entry.recv_neighbors) &&
+            IsSameList(entry.send_neighbors, new_entry.send_neighbors) &&
+            IsSameList(entry.send_weights, new_entry.send_weights) &&
+            IsSameList(entry.recv_neighbors, new_entry.recv_neighbors) &&
             tensor_size + new_tensor_size <= state.tensor_fusion_threshold) {
           // These tensors will fuse together well.
           tensor_size += new_tensor_size;
@@ -1020,6 +1037,7 @@ void NegotiateOfRequestOfMaster(BluefogGlobalState& state,
     } else {
       PerformOperation(nego_entries);
     }
+    state.unfinished_enqueued_entries.fetch_sub(nego_entries.size());
   }
 
   // Check for stalled tensors.
@@ -1034,6 +1052,7 @@ void NegotiateOfRequestOfSlave(BluefogGlobalState& state,
                                std::deque<Request>& message_queue_buffer,
                                bool& should_change_topo,
                                bool& should_shut_down) {
+  state.unfinished_enqueued_entries.fetch_add(message_queue_buffer.size());
   std::string encoded_message;
   RequestList message_list;
   message_list.set_shutdown(state.shut_down);
@@ -1042,6 +1061,7 @@ void NegotiateOfRequestOfSlave(BluefogGlobalState& state,
     message_list.add_request(message_queue_buffer.front());
     message_queue_buffer.pop_front();
   }
+
   RequestList::SerializeToString(message_list, encoded_message);
   int encoded_message_length = (int)encoded_message.length() + 1;
   MPI_Gather(&encoded_message_length, 1, MPI_INT, nullptr, 1, MPI_INT,
@@ -1069,6 +1089,7 @@ void NegotiateOfRequestOfSlave(BluefogGlobalState& state,
     } else {
       PerformOperation(nego_entries);
     }
+    state.unfinished_enqueued_entries.fetch_sub(nego_entries.size());
   }
 
   if (response_list.shutdown()) {
@@ -1082,6 +1103,8 @@ void NegotiateOfRequestOfSlave(BluefogGlobalState& state,
 void NegotiationOfRequest(BluefogGlobalState& state,
                           std::deque<Request>& message_queue_buffer,
                           bool& should_change_topo, bool& should_shut_down) {
+  // TODO(ybc) should_change_topo has no effect after condition variable refactor. 
+  // Just keep it for a while. will remove.
   if (bluefog_rank() == COORDINATE_RANK) {
     NegotiateOfRequestOfMaster(state, message_queue_buffer, should_change_topo,
                                should_shut_down);
@@ -1096,6 +1119,15 @@ bool RunLoopOnce(BluefogGlobalState& state) {
   bool should_shut_down = state.shut_down;
   bool should_change_topo = state.setting_topology;
 
+  std::unique_lock<std::mutex> lk(state.loop_mutex);
+  // The real mutex for queue is the on under TensorQueue.
+  state.loop_cv.wait_for(lk, std::chrono::seconds(10), [&state] {
+    // When we requesting shut_down, or any unfinished entries waiting in the
+    // negotiation we should not wait.
+    return state.shut_down || (state.unfinished_enqueued_entries > 0) ||
+           (state.tensor_queue.size() > 0);
+  });
+
   // This delay determines thread frequency and MPI message latency
   auto sleep_duration =
       state.last_cycle_start +
@@ -1105,6 +1137,10 @@ bool RunLoopOnce(BluefogGlobalState& state) {
     std::this_thread::sleep_for(sleep_duration);
   }
   state.last_cycle_start = std::chrono::steady_clock::now();
+  if (global_background_thread_suspend) {
+    BFLOG(ERROR) << "Suspending";
+    std::this_thread::sleep_for(std::chrono::seconds(3));
+  }
 
   std::deque<Request> message_queue_buffer;
   state.tensor_queue.PopMessagesFromQueue(message_queue_buffer);
@@ -1112,7 +1148,7 @@ bool RunLoopOnce(BluefogGlobalState& state) {
   std::vector<TensorTableEntry> entries;
   auto IsRequestConvertToEntryDirectly = [](const Request& request) -> bool {
     return global_skip_negotiate_stage ||
-           (request.request_type() != Request::ALLREDUCE &&
+            (request.request_type() != Request::ALLREDUCE &&
             request.request_type() != Request::ALLGATHER &&
             request.request_type() != Request::BROADCAST &&
             request.request_type() != Request::NEIGHBOR_ALLREDUCE &&
@@ -1131,7 +1167,8 @@ bool RunLoopOnce(BluefogGlobalState& state) {
       std::remove_if(message_queue_buffer.begin(), message_queue_buffer.end(),
                      IsRequestConvertToEntryDirectly),
       message_queue_buffer.end());
-
+   
+  lk.unlock();  // Never hold the mutex when there is remote function.
   PerformOperation(entries);
 
   // For the rest requests, they needs to coordinate and neogiate.
@@ -1144,19 +1181,6 @@ bool RunLoopOnce(BluefogGlobalState& state) {
     NegotiationOfRequest(state, message_queue_buffer, should_change_topo,
                          should_shut_down);
   }
-  // Seperate the setting topology and negotiate communnication.
-  if (should_change_topo) {
-    bluefog_global.ready_to_setting_topology = true;
-    while (!bluefog_global.setting_topology_done) {
-      std::this_thread::sleep_for(SUSPEND_BACKGROUND_WAITTING_DURATION);
-    }
-    bluefog_global.ready_to_setting_topology = false;
-    // Wait for main thread reset.
-    while (bluefog_global.setting_topology_done) {
-      std::this_thread::sleep_for(SUSPEND_BACKGROUND_WAITTING_DURATION);
-    }
-  }
-
   return !should_shut_down;
 }
 
@@ -1195,6 +1219,7 @@ void bluefog_init() { InitializeBluefogOnce(); }
 void bluefog_shutdown() {
   if (bluefog_global.background_thread.joinable()) {
     bluefog_global.shut_down = true;
+    bluefog_global.loop_cv.notify_all();
     bluefog_global.background_thread.join();
     // Reset the initialization flag to allow restarting with bluefog_init(...)
     //bluefog_global.initialize_flag.clear();
@@ -1270,36 +1295,21 @@ int bluefog_set_topology(int indegree, const int* sources, int outdegree,
     return -1;
   }
 #endif
-  bluefog_global.setting_topology = true;
-  while (!bluefog_global.ready_to_setting_topology.load()) {
-    std::this_thread::sleep_for(SUSPEND_BACKGROUND_WAITTING_DURATION);
-  }
-  bluefog_global.tensor_queue.LockTensorQueue();
-  if (bluefog_global.tensor_queue.size() > 0) {
-    BFLOG(ERROR)
-        << "Cannot set the topology because there are unfinished MPI ops.";
-    bluefog_global.tensor_queue.UnlockTensorQueue();
-    return -1;
-  }
-
-  bool mpi_result = bluefog_global.controller->SetTopology(
-      indegree, sources, outdegree, destinations);
+  bool mpi_result;
+  // When we change the topology, there should be no entries being processed at
+  // same time.
+  {
+    std::lock_guard<std::mutex> lk(bluefog_global.loop_mutex);
+    mpi_result = bluefog_global.controller->SetTopology(
+        indegree, sources, outdegree, destinations);
 #if HAVE_NCCL && NCCL_MINOR < 7
-  if (mpi_result && nccl_context.is_initialized) {
-    bluefog_global.nccl_controller->DestroyPeerCommunicators();
-    bluefog_global.nccl_controller->InitPeerCommunicators();
-  }
+    if (mpi_result && nccl_context.is_initialized) {
+      bluefog_global.nccl_controller->DestroyPeerCommunicators();
+      bluefog_global.nccl_controller->InitPeerCommunicators();
+    }
 #endif
-  bluefog_global.tensor_queue.UnlockTensorQueue();
-
-  bluefog_global.setting_topology = false;
-  bluefog_global.setting_topology_done = true;
-  // Wait for the background thread receive the setting_topology_done and
-  // close the ready_to_setting_topology epoch.
-  while (bluefog_global.ready_to_setting_topology) {
-    std::this_thread::sleep_for(SUSPEND_BACKGROUND_WAITTING_DURATION);
   }
-  bluefog_global.setting_topology_done = false;
+  bluefog_global.loop_cv.notify_one();
   return mpi_result;
 }
 
@@ -1379,6 +1389,16 @@ int bluefog_get_skip_negotiate_stage() {
   return GetSkipNegotiateStageState();
 }
 
+int bluefog_suspend() {
+  global_background_thread_suspend = true;
+  return 1;
+}
+
+int bluefog_resume() {
+  global_background_thread_suspend = false;
+  return 1;
+}
+
 }  // extern "C"
 
 Status EnqueueTensorAllreduce(std::shared_ptr<Tensor> tensor,
@@ -1413,7 +1433,14 @@ Status EnqueueTensorAllreduce(std::shared_ptr<Tensor> tensor,
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.tensor_queue.AddToTensorQueue(e, message);
+  {
+    std::unique_lock<std::mutex> lk(bluefog_global.loop_mutex);
+    bluefog_global.loop_cv.notify_one();
+  }
   return status;
 }
 
@@ -1445,7 +1472,14 @@ Status EnqueueTensorBroadcast(std::shared_ptr<Tensor> tensor,
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.tensor_queue.AddToTensorQueue(e, message);
+  {
+    std::unique_lock<std::mutex> lk(bluefog_global.loop_mutex);
+    bluefog_global.loop_cv.notify_one();
+  }
   return status;
 }
 
@@ -1476,13 +1510,24 @@ Status EnqueueTensorAllgather(std::shared_ptr<Tensor> tensor,
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.tensor_queue.AddToTensorQueue(e, message);
+  {
+    std::unique_lock<std::mutex> lk(bluefog_global.loop_mutex);
+    bluefog_global.loop_cv.notify_one();
+  }
   return status;
 }
 
 Status EnqueueTensorNeighborAllgather(std::shared_ptr<Tensor> tensor,
                                       std::shared_ptr<OpContext> context,
                                       std::shared_ptr<ReadyEvent> ready_event,
+                                      const std::vector<int>& src_ranks,
+                                      const std::vector<int>& dst_ranks,
+                                      bool dynamic_neighbors_enabled,
+                                      bool enable_topo_check,
                                       const std::string& name, const int device,
                                       StatusCallback callback) {
   Request message;
@@ -1491,6 +1536,8 @@ Status EnqueueTensorNeighborAllgather(std::shared_ptr<Tensor> tensor,
   message.set_tensor_type(tensor->dtype());
   message.set_device(device);
   message.set_request_type(Request::NEIGHBOR_ALLGATHER);
+  // TODO(ybc) Add src_ranks and dst_ranks to messages?
+  // Probably no need since we do not fuse neighbor_allgather.
   for (int i = 0; i < tensor->shape().dims(); i++) {
     message.add_tensor_shape((int64_t)tensor->shape().dim_size(i));
   }
@@ -1500,6 +1547,10 @@ Status EnqueueTensorNeighborAllgather(std::shared_ptr<Tensor> tensor,
   e.tensor = tensor;
   e.context = context;
   e.device = device;
+  e.dynamic_neighbors_enabled = dynamic_neighbors_enabled;
+  e.enable_topo_check = enable_topo_check;
+  e.send_neighbors = std::make_shared<std::vector<int>>(dst_ranks);
+  e.recv_neighbors = std::make_shared<std::vector<int>>(src_ranks);
   e.ready_event = ready_event;
   e.callback = callback;
   e.mpi_ops_type = MPIOpsType::NEIGHBOR_ALLGATHER;
@@ -1507,7 +1558,14 @@ Status EnqueueTensorNeighborAllgather(std::shared_ptr<Tensor> tensor,
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.tensor_queue.AddToTensorQueue(e, message);
+  {
+    std::unique_lock<std::mutex> lk(bluefog_global.loop_mutex);
+    bluefog_global.loop_cv.notify_one();
+  }
   return status;
 }
 
@@ -1517,7 +1575,9 @@ Status EnqueueTensorNeighborAllreduce(std::shared_ptr<Tensor> tensor,
                                       std::shared_ptr<ReadyEvent> ready_event,
                                       std::shared_ptr<std::vector<int>> recv_neighbors,
                                       std::shared_ptr<std::vector<int>> send_neighbors,
+                                      std::shared_ptr<std::vector<double>> send_weights,
                                       bool dynamic_neighbors_enabled,
+                                      bool dst_weighting_enabled,
                                       bool is_hierarchical,
                                       bool enable_topo_check,
                                       const std::string& name, const int device,
@@ -1541,7 +1601,9 @@ Status EnqueueTensorNeighborAllreduce(std::shared_ptr<Tensor> tensor,
   e.ready_event = ready_event;
   e.recv_neighbors = recv_neighbors;
   e.send_neighbors = send_neighbors;
+  e.send_weights = send_weights;
   e.dynamic_neighbors_enabled = dynamic_neighbors_enabled;
+  e.dst_weighting_enabled = dst_weighting_enabled;
   e.is_hierarchical = is_hierarchical;
   e.enable_topo_check = enable_topo_check;
   e.device = device;
@@ -1551,7 +1613,14 @@ Status EnqueueTensorNeighborAllreduce(std::shared_ptr<Tensor> tensor,
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.tensor_queue.AddToTensorQueue(e, message);
+  {
+    std::unique_lock<std::mutex> lk(bluefog_global.loop_mutex);
+    bluefog_global.loop_cv.notify_one();
+  }
   return status;
 }
 
@@ -1588,7 +1657,14 @@ Status EnqueueTensorPairGossip(std::shared_ptr<Tensor> tensor,
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.tensor_queue.AddToTensorQueue(e, message);
+  {
+    std::unique_lock<std::mutex> lk(bluefog_global.loop_mutex);
+    bluefog_global.loop_cv.notify_one();
+  }
   return status;
 }
 
@@ -1617,7 +1693,14 @@ Status EnqueueTensorWindowCreate(
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.tensor_queue.AddToTensorQueue(e, message);
+  {
+    std::unique_lock<std::mutex> lk(bluefog_global.loop_mutex);
+    bluefog_global.loop_cv.notify_one();
+  }
   return status;
 }
 
@@ -1638,7 +1721,14 @@ Status EnqueueTensorWindowFree(const std::string& name, int device,
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.tensor_queue.AddToTensorQueue(e, message);
+  {
+    std::unique_lock<std::mutex> lk(bluefog_global.loop_mutex);
+    bluefog_global.loop_cv.notify_one();
+  }
   return status;
 }
 
@@ -1668,7 +1758,14 @@ Status EnqueueTensorWindowPut(std::shared_ptr<Tensor> tensor,
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.tensor_queue.AddToTensorQueue(e, message);
+  {
+    std::unique_lock<std::mutex> lk(bluefog_global.loop_mutex);
+    bluefog_global.loop_cv.notify_one();
+  }
   return status;
 }
 
@@ -1696,7 +1793,14 @@ Status EnqueueTensorWindowAccumulate(
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.tensor_queue.AddToTensorQueue(e, message);
+  {
+    std::unique_lock<std::mutex> lk(bluefog_global.loop_mutex);
+    bluefog_global.loop_cv.notify_one();
+  }
   return status;
 }
 
@@ -1720,7 +1824,14 @@ Status EnqueueTensorWindowGet(const std::string& name,
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.tensor_queue.AddToTensorQueue(e, message);
+  {
+    std::unique_lock<std::mutex> lk(bluefog_global.loop_mutex);
+    bluefog_global.loop_cv.notify_one();
+  }
   return status;
 }
 
@@ -1733,6 +1844,9 @@ Status ExecuteBarrier(StatusCallback callback) {
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   bluefog_global.controller->Barrier(e);
   return Status::OK();
 }
@@ -1741,6 +1855,9 @@ Status WindowSync(const std::string& name, int device) {
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Vendor vendor = DetermineController(MPIOpsType::WIN_SYNC, device);
   Status status;
 #if HAVE_NCCL
@@ -1767,6 +1884,9 @@ Status WindowMutexAcquire(const std::string& name,
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   // Because mutex is always associated with each win_create ops, it is safe to
   // use the vendor of win_create for window mutex.
   Vendor vendor = DetermineController(MPIOpsType::WIN_CREATE, device);
@@ -1795,6 +1915,9 @@ Status WindowMutexRelease(const std::string& name,
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   // Because mutex is always associated with each win_create ops, it is safe to
   // use the vendor of win_create for window mutex.
   Vendor vendor = DetermineController(MPIOpsType::WIN_CREATE, device);
@@ -1822,6 +1945,9 @@ Status GetBluefogTimeline(Timeline*& timeline) {
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   if (!bluefog_global.timeline_enabled) {
     return Status::Aborted("timeline is not enabled.");
   }
@@ -1833,6 +1959,9 @@ Status GetBluefogFusionBuffer(FusionBufferManager*& fusion_buffer) {
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   return Status::OK();
 }
 
@@ -1841,6 +1970,9 @@ Status WindowFence(const std::string& name) {
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.controller->WinFence(name);
 
   if (!status.ok()) {
@@ -1854,6 +1986,9 @@ Status WindowLock(const std::string& name) {
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.controller->WinLock(name);
 
   if (!status.ok()) {
@@ -1867,6 +2002,9 @@ Status WindowUnlock(const std::string& name) {
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   Status status = bluefog_global.controller->WinUnlock(name);
 
   if (!status.ok()) {
@@ -1880,6 +2018,9 @@ Status GetWindowVersion(const std::string& name, std::vector<int>& versions) {
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
 
   Status status =
       bluefog_global.controller->GetWindowVersionValue(name, versions);
@@ -1897,6 +2038,9 @@ Status GetWinAssociatedPByNameAndRank(const std::string& name,
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   return bluefog_global.controller->GetWinAssociatedPByNameAndRank(
       name, rank, weight);
 }
@@ -1906,6 +2050,9 @@ Status SetWinAssociatedPByNameAndRank(const std::string& name,
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
+  if (global_background_thread_suspend) {
+    return SUSPEND_ERROR;
+  }
   return bluefog_global.controller->SetWinAssociatedPByNameAndRank(
       name, rank, weight);
 }
@@ -1927,27 +2074,14 @@ void SetSkipNegotiateStageState(bool value) {
   if (value == global_skip_negotiate_stage) {
     return;
   }
-  if (value) {
-    // From running negotiate to skipping negotiate, we need to properly turn
-    // off negotiate stage. Otherwise, it may hang the processes. Use setting
-    // topology flag to suspend the negotiate stage then skip it.
-    bluefog_global.setting_topology = true;
-    while (!bluefog_global.ready_to_setting_topology.load()) {
-      std::this_thread::sleep_for(SUSPEND_BACKGROUND_WAITTING_DURATION);
-    }
 
+  // From running negotiate to skipping negotiate, we need to properly turn
+  // off negotiate stage. Otherwise, it may hang the processes. Use setting
+  // topology flag to suspend the negotiate stage then skip it.
+  {
+    std::lock_guard<std::mutex> lk(bluefog_global.loop_mutex);
     global_skip_negotiate_stage = value;
-
-    bluefog_global.setting_topology = false;
-    bluefog_global.setting_topology_done = true;
-    // Wait for the background thread receive the setting_topology_done and
-    // close the ready_to_setting_topology epoch.
-    while (bluefog_global.ready_to_setting_topology) {
-      std::this_thread::sleep_for(SUSPEND_BACKGROUND_WAITTING_DURATION);
-    }
-    bluefog_global.setting_topology_done = false;
-  } else {
-    global_skip_negotiate_stage = value;
+    bluefog_global.loop_cv.notify_one();
   }
 }
 
