@@ -92,7 +92,9 @@ ALLGATHER_H(torch_cuda_DoubleTensor, THCudaDoubleTensor)
 
 #define NEIGHBOR_ALLGATHER_H(torch_Tensor, THTensor)                              \
   extern "C" int bluefog_torch_neighbor_allgather_nonblocking_##torch_Tensor(     \
-      THTensor* tensor, THTensor* output, char* name);
+      THTensor* tensor, THTensor* output, const std::vector<int>& src_ranks,      \
+      const std::vector<int>& dst_ranks, bool dynamic_neighbors_enabled,          \
+      bool enable_topo_check, char* name);
 
 NEIGHBOR_ALLGATHER_H(torch_ByteTensor, THByteTensor)
 NEIGHBOR_ALLGATHER_H(torch_CharTensor, THCharTensor)
@@ -111,12 +113,13 @@ NEIGHBOR_ALLGATHER_H(torch_cuda_FloatTensor, THCudaTensor)
 NEIGHBOR_ALLGATHER_H(torch_cuda_DoubleTensor, THCudaDoubleTensor)
 #endif
 
-#define NEIGHBOR_ALLREDUCE_H(torch_Tensor, THTensor)                             \
-  extern "C" int bluefog_torch_neighbor_allreduce_nonblocking_##torch_Tensor(    \
-      THTensor* tensor, THTensor* output, double self_weight,                    \
-      const std::unordered_map<int, double>& neighbor_weights,                   \
-      const std::vector<int>& send_neighbors, bool dynamic_neighbors_enabled,    \
-      bool enable_topo_check, bool avg_computation, bool is_hierarchical,        \
+#define NEIGHBOR_ALLREDUCE_H(torch_Tensor, THTensor)                           \
+  extern "C" int bluefog_torch_neighbor_allreduce_nonblocking_##torch_Tensor(  \
+      THTensor* tensor, THTensor* output, double self_weight,                  \
+      const std::unordered_map<int, double>& src_weights,                      \
+      const std::unordered_map<int, double>& dst_weights,                      \
+      bool dynamic_neighbors_enabled, bool dst_weighting_enabled,              \
+      bool enable_topo_check, bool avg_computation, bool is_hierarchical,      \
       char* name);
 
 NEIGHBOR_ALLREDUCE_H(torch_HalfTensor, THHalfTensor)
