@@ -15,11 +15,13 @@
 # ==============================================================================
 
 from contextlib import contextmanager
-from typing import List, Dict, Optional
+from typing import List, Dict, Union, Optional
 
+import numpy as np
 import torch
 
 from bluefog.torch import mpi_lib  # C library
+from bluefog.torch.utility import deprecated_function_arg
 from bluefog.common.basics import BlueFogBasics, logger
 from bluefog.common.topology_util import GetRecvWeights
 
@@ -29,14 +31,21 @@ _basics = BlueFogBasics(__file__, 'mpi_lib')
 init = _basics.init
 shutdown = _basics.shutdown
 size = _basics.size
+machine_size = _basics.machine_size
 local_size = _basics.local_size
 rank = _basics.rank
+machine_rank = _basics.machine_rank
 local_rank = _basics.local_rank
 load_topology = _basics.load_topology
+load_machine_topology = _basics.load_machine_topology
 is_topo_weighted = _basics.is_topo_weighted
+is_machine_topo_weighted = _basics.is_machine_topo_weighted
 set_topology = _basics.set_topology
+set_machine_topology = _basics.set_machine_topology
 in_neighbor_ranks = _basics.in_neighbor_ranks
+in_neighbor_machine_ranks = _basics.in_neighbor_machine_ranks
 out_neighbor_ranks = _basics.out_neighbor_ranks
+out_neighbor_machine_ranks = _basics.out_neighbor_machine_ranks
 mpi_threads_supported = _basics.mpi_threads_supported
 unified_mpi_window_model_supported = _basics.unified_mpi_window_model_supported
 is_homogeneous = _basics.is_homogeneous
@@ -48,6 +57,9 @@ timeline_context = _basics.timeline_context
 timeline_start_activity = _basics.timeline_start_activity
 timeline_end_activity = _basics.timeline_end_activity
 
+suspend = _basics.suspend
+resume = _basics.resume
+
 # Schema: handle -> input, output
 # We keep input in order to make sure it does not get garbage collected
 # before the operation is finished.
@@ -365,123 +377,164 @@ def _neighbor_allgather_function_factory(tensor):
     return 'bluefog_torch_neighbor_allgather_nonblocking_' + tensor.type().replace('.', '_')
 
 
-def _neighbor_allgather_nonblocking(tensor, output, name):
+def _neighbor_allgather_nonblocking(tensor, output, src_ranks, dst_ranks, enable_topo_check, name):
     function = _check_function(_neighbor_allgather_function_factory, tensor)
-    handle = getattr(mpi_lib, function)(tensor, output,
+    if dst_ranks is None and src_ranks is None:
+        src_ranks, dst_ranks = [], []
+        dynamic_neighbors_enabled = False
+    elif dst_ranks is not None and src_ranks is not None:
+        dynamic_neighbors_enabled = True
+    else:
+        raise ValueError(
+            "Argument dst_ranks and src_ranks have to be presented at the same time")
+    handle = getattr(mpi_lib, function)(tensor, output, src_ranks, dst_ranks,
+                                        dynamic_neighbors_enabled, enable_topo_check,
                                         name.encode() if name is not None else "")
     _handle_map[handle] = (tensor, output)
     return handle
 
 
-def neighbor_allgather(tensor: torch.Tensor, name: Optional[str] = None) -> torch.Tensor:
+def neighbor_allgather(tensor: torch.Tensor, *,
+                       src_ranks: Optional[List] = None,
+                       dst_ranks: Optional[List] = None,
+                       enable_topo_check: bool = True,
+                       name: Optional[str] = None) -> torch.Tensor:
     """
     A function that concatenates the input tensor with the same input tensor on
     on all neighbor Bluefog processes (Not include self). The input tensor is not modified.
 
     The concatenation is done on the first dimension, so the input tensors on the
-    different processes must have the same rank and shape.
+    different processes must have the same shape except the first dimension.
+    For example: rank 0 with tensor shape [3, 5, 4] and rank 1 with tensor shape [5, 5, 4] are
+    allowed, the output will be [8, 5, 4] assuming two are connected.
+
+    If src_ranks and dst_ranks are not provided, the neighbor_allgather gathers the tensors according
+    to the global default topology, which can be obtained through `bf.in_neighbor_ranks`.
+    Otherwise, the order is the same as src_ranks provided.
 
     Arguments:
         tensor: A tensor to allgather.
+        dst_ranks: A list of destination ranks. If present, ignoring global topology setting.
+            This argument is useful under dynamic topology case.
+            Note dst_ranks and src_ranks should be presented at same time and compatible.
+        src_ranks: A list of source ranks.
+        enable_topo_check: When send_neighbors is present, enabling this option checks if the
+            sending and recieving neighbors match with each other. Disabling this check can boost
+            the performance.
         name: A name of the allgather operation.
 
     Returns:
-        A tensor of the same type as `tensor`, concatenated on dimension zero
-        across all processes. The shape is identical to the input shape, except for
-        the first dimension, which may be greater and is the sum of all first
-        dimensions of the tensors in neighbor Bluefog processes.
+        A tensor of the same type as `tensor`, concatenated on dimension zero.
+        The shape is identical to the input shape, except for
+        the first dimension. The order of gathered tensors is guaranteed to be the same order as
+        the src_ranks if specified or in_neighbor_ranks through the default topology.
     """
-    handle = neighbor_allgather_nonblocking(tensor, name)
+    handle = neighbor_allgather_nonblocking(
+        tensor, src_ranks=src_ranks, dst_ranks=dst_ranks, enable_topo_check=enable_topo_check, name=name)
     return synchronize(handle)
 
 
-def neighbor_allgather_nonblocking(tensor: torch.Tensor, name: Optional[str] = None) -> int:
+def neighbor_allgather_nonblocking(tensor: torch.Tensor, *,
+                                   src_ranks: Optional[List] = None,
+                                   dst_ranks: Optional[List] = None,
+                                   enable_topo_check: bool = True,
+                                   name: Optional[str] = None) -> int:
     """
     A function that nonblockingly concatenates the input tensor with the same input
     tensor on all neighbor Bluefog processes (Not include self).
     The input tensor is not modified.
 
     The concatenation is done on the first dimension, so the input tensors on the
-    different processes must have the same rank and shape.
+    different processes must have the same shape except the first dimension.
+    For example: rank 0 with tensor shape [3, 5, 4] and rank 1 with tensor shape [5, 5, 4] are
+    allowed, the output will be [8, 5, 4] assuming two are connected.
+
+    If src_ranks and dst_ranks is not provided, the neighbor_allgather gather the tensors according
+    to the global default topology.
 
     Arguments:
         tensor: A tensor to allgather.
+        src_ranks: A list of source ranks. If present, ignoring global topology setting.
+            This argument is useful under dynamic topology case.
+            Note dst_ranks and src_ranks should be presented at same time and compatible.
+        dst_ranks: A list of destination ranks.
+        enable_topo_check: When send_neighbors is present, enabling this option checks if the
+            sending and recieving neighbors match with each other. Disabling this check can boost
+            the performance.
         name: A name of the allgather operation.
 
     Returns:
-        A handle to the allgather operation that can be used with `poll()` or
-        `synchronize()`.
+        A handle to the neighbor_allgather operation that can be used with `poll()` or
+        `synchronize()`. Check neighbor_allgather function for the output tensor information.
     """
     output = tensor.new()  # real size will be allocated later.
-    return _neighbor_allgather_nonblocking(tensor, output, name)
+    return _neighbor_allgather_nonblocking(tensor, output, src_ranks, dst_ranks,
+                                           enable_topo_check, name)
 
 
 def _neighbor_allreduce_function_factory(tensor):
     return 'bluefog_torch_neighbor_allreduce_nonblocking_' + tensor.type().replace('.', '_')
 
 
-def _neighbor_allreduce_nonblocking(tensor, output, self_weight, neighbor_weights,
-                                    send_neighbors, enable_topo_check, name):
+def _neighbor_allreduce_nonblocking(tensor, output, self_weight, src_weights,
+                                    dst_weights, enable_topo_check, name):
     function = _check_function(_neighbor_allreduce_function_factory, tensor)
-    if send_neighbors is None:
-        send_neighbors = []
+    if dst_weights is None:
+        dst_weights = {}
         dynamic_neighbors_enabled = False
-    elif len(set(send_neighbors)) != len(send_neighbors):
-        raise ValueError("Argument send_neighbors should only contain the unique ranks.")
-    elif self_weight is None or neighbor_weights is None:
-        raise ValueError("Arguments self_weight and neighbor_weights should be presented if "
+        dst_weighting_enabled = False
+    elif len(set(dst_weights)) != len(dst_weights):
+        raise ValueError("Argument dst_weights should only contain the unique ranks.")
+    elif self_weight is None or src_weights is None:
+        raise ValueError("Arguments self_weight and src_weights should be presented if "
                          "enabling dynamic topology.")
-    elif not send_neighbors:
-        raise ValueError("Argument send_neighbors cannot be empty list but we plan to support "
-                         "it in future.")
     else:
         dynamic_neighbors_enabled = True
-    if self_weight is None and neighbor_weights is None:
+        if isinstance(dst_weights, list):
+            dst_weights = {dst:1.0 for dst in dst_weights}
+        dst_weighting_enabled = not np.allclose(list(dst_weights.values()), 1.0)
+    if self_weight is None and src_weights is None:
         # Implying this is static graph.
         if is_topo_weighted():
             topology = load_topology()
-            self_weight, neighbor_weights = GetRecvWeights(topology, rank())
-            avg_computation = True
+            self_weight, src_weights = GetRecvWeights(topology, rank())
+            weighted_average_computation = True
         else:
             weight = 1.0/(len(in_neighbor_ranks())+1)
             self_weight = weight
-            neighbor_weights = {r: weight for r in in_neighbor_ranks()}
-            avg_computation = False
-    elif self_weight is not None and neighbor_weights is not None:
-        if not isinstance(neighbor_weights, dict):
+            src_weights = {r: weight for r in in_neighbor_ranks()}
+            weighted_average_computation = False
+    elif self_weight is not None and src_weights is not None:
+        if not isinstance(src_weights, dict):
             raise ValueError("Argument neighbor_weights has to be a dictionary map from the "
                              "(in-)neighbor rank to the weights.")
         if not isinstance(self_weight, float):
             raise ValueError(
                 "Argument self_weight has to be a float for self rank.")
         if not dynamic_neighbors_enabled and \
-           not set(neighbor_weights.keys()).issubset(set(in_neighbor_ranks())):
+           not set(src_weights.keys()).issubset(set(in_neighbor_ranks())):
             raise ValueError("The key of weights should only contain the ranks that belong to "
                              " in-neighbors and self rank.")
-        uniform_weights = 1.0/(len(neighbor_weights)+1)
-        avg_computation = False
-        if abs(self_weight - uniform_weights) > 1e-6:
-            avg_computation = True
-        for n_weights in neighbor_weights.values():
-            if abs(n_weights - uniform_weights) > 1e-6:
-                avg_computation = True
-                break
+        uniform_weights = 1.0/(len(src_weights)+1)
+        weighted_average_computation = not(np.isclose(self_weight, uniform_weights) and
+                                           np.allclose(list(src_weights.values()), uniform_weights))
     else:
         raise ValueError("Arguments self_weight and neighbor_weights have to be presented at "
                          "the same time")
     is_hierarchical = False
-    handle = getattr(mpi_lib, function)(tensor, output, self_weight, neighbor_weights,
-                                        send_neighbors, dynamic_neighbors_enabled,
-                                        enable_topo_check, avg_computation,
+    handle = getattr(mpi_lib, function)(tensor, output, self_weight, src_weights, dst_weights,
+                                        dynamic_neighbors_enabled, dst_weighting_enabled,
+                                        enable_topo_check, weighted_average_computation,
                                         is_hierarchical, name.encode() if name is not None else "")
     _handle_map[handle] = (tensor, output)
     return handle
 
-
-def neighbor_allreduce(tensor: torch.Tensor,
+@deprecated_function_arg(arg_name="neighbor_weights", fix="Use src_weights instead")
+@deprecated_function_arg(arg_name="send_neighbors", fix="Use dst_weights instead")
+def neighbor_allreduce(tensor: torch.Tensor, *,
                        self_weight: Optional[float] = None,
-                       neighbor_weights: Optional[Dict[int, float]] = None,
-                       send_neighbors: Optional[List[int]] = None,
+                       src_weights: Optional[Dict[int, float]] = None,
+                       dst_weights: Optional[Union[Dict[int, float], List[int]]] = None,
                        enable_topo_check: bool = True,
                        name: Optional[str] = None) -> torch.Tensor:
     """
@@ -497,19 +550,19 @@ def neighbor_allreduce(tensor: torch.Tensor,
 
     Arguments:
         tensor: A tensor to execute weighted average with neighbors.
-        self_weight: The weight for self node, used with neighbor_weights.
-        neighbor_weights: The weights for in-neighbor nodes, used with self weight.
-            If neighbor_weights is presented, the return tensor will return the weighted average
+        self_weight: The weight for self node, used with src_weights.
+        src_weights: The weights for in-neighbor nodes, used with self weight.
+            If src_weights is presented, the return tensor will return the weighted average
             defined by these weights and the self_weight. If not, the return tensor will return
             the weighted average defined by the topology weights is provided or uniformly average.
             The data structure of weights should be {rank : weight} and rank has to belong to the
             (in-)neighbors.
-        send_neighbors: The list of neighbor nodes to be sent to. If set to be None, assume the
+        dst_weights: The weights for out-neighbor nodes. If set to be None, assume the
             the current node sends to all of its (out-)neighbors. If having values, assume only
-            part of (out-)neighbors will be sent to. In this mode, this node sends its value to
-            partial neighbors listed in this variable in a dynamic graph, and `self_weight` and
-            `neighbor_weights` must be present.
-        enable_topo_check: When send_neighbors is present, enabling this option checks if the
+            part of (out-)neighbors will be sent to. If set to be a list, assume all the weights
+            are one. In this mode, this node sends its value to partial neighbors listed in this
+            variable in a dynamic graph, and `self_weight` and `src_weights` must be present.
+        enable_topo_check: When dst_weights is present, enabling this option checks if the
             sending and recieving neighbors match with each other. Disabling this check can boost
             the performance.
         name: A name of the reduction operation.
@@ -519,19 +572,26 @@ def neighbor_allreduce(tensor: torch.Tensor,
 
     Note: self_weight and neighbor_weights must be presented at the same time.
     """
-    if (self_weight is None and neighbor_weights is not None) or \
-       (self_weight is not None and neighbor_weights is None):
-        raise ValueError("Arguments self_weight and neighbor_weights have to be presented at "
+    # TODO(hanbinhu) #82 Symmetrical argument for self_weight, src_weights, dst_weights
+    if (self_weight is None and src_weights is not None) or \
+       (self_weight is not None and src_weights is None):
+        raise ValueError("Arguments self_weight and src_weights have to be presented at "
                          "the same time")
-    handle = neighbor_allreduce_nonblocking(tensor, self_weight, neighbor_weights,
-                                            send_neighbors, enable_topo_check, name)
+    handle = neighbor_allreduce_nonblocking(tensor,
+                                            self_weight=self_weight,
+                                            src_weights=src_weights,
+                                            dst_weights=dst_weights,
+                                            enable_topo_check=enable_topo_check,
+                                            name=name)
     return synchronize(handle)
 
 
-def neighbor_allreduce_nonblocking(tensor: torch.Tensor,
+@deprecated_function_arg(arg_name="neighbor_weights", fix="Use src_weights instead")
+@deprecated_function_arg(arg_name="send_neighbors", fix="Use dst_weights instead")
+def neighbor_allreduce_nonblocking(tensor: torch.Tensor, *,
                                    self_weight: Optional[float] = None,
-                                   neighbor_weights: Optional[Dict[int, float]] = None,
-                                   send_neighbors: Optional[List[int]] = None,
+                                   src_weights: Optional[Dict[int, float]] = None,
+                                   dst_weights: Optional[Union[Dict[int, float], List[int]]] = None,
                                    enable_topo_check: bool = True,
                                    name: Optional[str] = None) -> int:
     """
@@ -547,19 +607,19 @@ def neighbor_allreduce_nonblocking(tensor: torch.Tensor,
 
     Arguments:
         tensor: A tensor to execute weighted average with neighbors.
-        self_weight: The weight for self node, used with neighbor_weights.
-        neighbor_weights: The weights for in-neighbor nodes, used with self weight.
-            If neighbor_weights is presented, the return tensor will return the weighted average
+        self_weight: The weight for self node, used with src_weights.
+        src_weights: The weights for in-neighbor nodes, used with self weight.
+            If src_weights is presented, the return tensor will return the weighted average
             defined by these weights and the self_weight. If not, the return tensor will return
             the weighted average defined by the topology weights is provided or uniformly average.
             The data structure of weights should be {rank : weight} and rank has to belong to the
             (in-)neighbors.
-        send_neighbors: The list of neighbor nodes to be sent to. If set to be None, assume the
+        dst_weights: The weights for out-neighbor nodes. If set to be None, assume the
             the current node sends to all of its (out-)neighbors. If having values, assume only
-            part of (out-)neighbors will be sent to. In this mode, this node sends its value to
-            partial neighbors listed in this variable in a dynamic graph, and `self_weight` and
-            `neighbor_weights` must be present.
-        enable_topo_check: When send_neighbors is present, enabling this option checks if the
+            part of (out-)neighbors will be sent to. If set to be a list, assume all the weights
+            are one. In this mode, this node sends its value to partial neighbors listed in this
+            variable in a dynamic graph, and `self_weight` and `src_weights` must be present.
+        enable_topo_check: When dst_weights is present, enabling this option checks if the
             sending and recieving neighbors match with each other. Disabling this check can boost
             the performance.
         name: A name of the neighbor_allreduce operation.
@@ -570,24 +630,25 @@ def neighbor_allreduce_nonblocking(tensor: torch.Tensor,
 
     Note: self_weight and neighbor_weights must be presented at the same time.
     """
-    if (self_weight is None and neighbor_weights is not None) or \
-       (self_weight is not None and neighbor_weights is None):
-        raise ValueError("Arguments self_weight and neighbor_weights have to be presented at "
+    # TODO(hanbinhu) #82 Symmetrical argument for self_weight, src_weights, dst_weights
+    if (self_weight is None and src_weights is not None) or \
+       (self_weight is not None and src_weights is None):
+        raise ValueError("Arguments self_weight and src_weights have to be presented at "
                          "the same time")
-    if send_neighbors is None:
+    if dst_weights is None:
         first_dim = tensor.shape[0] * len(in_neighbor_ranks())
     else:
-        first_dim = tensor.shape[0] * len(neighbor_weights)
+        first_dim = tensor.shape[0] * len(src_weights)
     new_shape = torch.Size([first_dim] + list(tensor.shape[1:]))
     output = tensor.new(new_shape)  # Pre-allocate the memory for the output.
-    return _neighbor_allreduce_nonblocking(tensor, output, self_weight, neighbor_weights,
-                                           send_neighbors, enable_topo_check, name=name)
-
+    return _neighbor_allreduce_nonblocking(tensor, output, self_weight, src_weights,
+                                           dst_weights, enable_topo_check, name=name)
 
+# TODO(hanbinhu) #81 Add dst_weight for hierarchical neighbor allreduce.
 def hierarchical_neighbor_allreduce(tensor: torch.Tensor,
-                                    self_weight: float,
-                                    neighbor_machine_weights: Dict[int, float],
-                                    send_neighbor_machines: List[int],
+                                    self_weight: float = None,
+                                    neighbor_machine_weights: Dict[int, float] = None,
+                                    send_neighbor_machines: List[int] = None,
                                     enable_topo_check: bool = False,
                                     name: Optional[str] = None) -> torch.Tensor:
     """
@@ -636,9 +697,9 @@ def hierarchical_neighbor_allreduce(tensor: torch.Tensor,
 
 
 def hierarchical_neighbor_allreduce_nonblocking(tensor: torch.Tensor,
-                                                self_weight: float,
-                                                neighbor_machine_weights: Dict[int, float],
-                                                send_neighbor_machines: List[int],
+                                                self_weight: float = None,
+                                                neighbor_machine_weights: Dict[int, float] = None,
+                                                send_neighbor_machines: List[int] = None,
                                                 enable_topo_check: bool = False,
                                                 name: Optional[str] = None) -> int:
     """
@@ -674,14 +735,15 @@ def hierarchical_neighbor_allreduce_nonblocking(tensor: torch.Tensor,
         A handle to the hierarchical_neighbor_allreduce operation that can be used with `poll()` or
         `synchronize()`.
     """
-    if (self_weight is None or neighbor_machine_weights is None):
-        raise ValueError("Arguments self_weight and neighbor_weights cannot be empty or None.")
     if (self_weight is None and neighbor_machine_weights is not None) or \
        (self_weight is not None and neighbor_machine_weights is None):
         raise ValueError("Arguments self_weight and neighbor_weights have to be presented at "
                          "the same time")
 
-    first_dim = tensor.shape[0] * len(neighbor_machine_weights)
+    if send_neighbor_machines is None:
+        first_dim = tensor.shape[0] * len(in_neighbor_machine_ranks())
+    else:
+        first_dim = tensor.shape[0] * len(neighbor_machine_weights)
     new_shape = torch.Size([first_dim] + list(tensor.shape[1:]))
     output = tensor.new(new_shape)  # Pre-allocate the memory for the output.
 
@@ -697,7 +759,22 @@ def _hierarchical_neighbor_allreduce_nonblocking(
         "hierarchical_neighbor_allreduce should be used under homogeneous environment only"
     assert local_size() > 1, "If local size is 1, you should use neighbor allreduce directly."
     function = _check_function(_neighbor_allreduce_function_factory, tensor)
-    if self_weight is not None and neighbor_machine_weights is not None:
+    if self_weight is None and neighbor_machine_weights is None and send_neighbor_machines is None:
+        # Implying this is static machine graph.
+        topology = load_machine_topology()
+        if topology is None:
+            raise RuntimeError("Machine topology must be set before the use of hierarchical "
+                               "neighbor allreduce")
+        if is_machine_topo_weighted():
+            self_weight, neighbor_machine_weights = GetRecvWeights(topology, machine_rank())
+            weighted_average_computation = True
+        else:
+            weight = 1.0/(len(in_neighbor_machine_ranks())+1)
+            self_weight = weight
+            neighbor_machine_weights = {r: weight for r in in_neighbor_machine_ranks()}
+            weighted_average_computation = False
+        send_neighbor_machines = out_neighbor_machine_ranks()
+    elif self_weight is not None and neighbor_machine_weights is not None:
         if not isinstance(neighbor_machine_weights, dict):
             raise ValueError("Argument neighbor_weights has to be a dictionary map from the "
                              "(in-)neighbor rank to the weights.")
@@ -705,18 +782,19 @@ def _hierarchical_neighbor_allreduce_nonblocking(
             raise ValueError(
                 "Argument self_weight has to be a float for self rank.")
         uniform_weights = 1.0/(len(neighbor_machine_weights)+1)
-        avg_computation = False
+        weighted_average_computation = False
         if abs(self_weight - uniform_weights) > 1e-6:
-            avg_computation = True
+            weighted_average_computation = True
         for n_weights in neighbor_machine_weights.values():
             if abs(n_weights - uniform_weights) > 1e-6:
-                avg_computation = True
+                weighted_average_computation = True
                 break
+        if not send_neighbor_machines:
+            raise ValueError("Argument send_neighbor_machines has to be presented and non-empty.")
     else:
-        raise ValueError("Arguments self_weight and neighbor_weights cannot be empty or None.")
+        raise ValueError("Arguments self_weight and neighbor_weights have to be presented at the "
+                         "same time.")
 
-    if not send_neighbor_machines:
-        raise ValueError("Argument send_neighbor_machines has to be presented and non-empty.")
 
     machine_size = size() // local_size()
     # Translate machine id into rank id.
@@ -735,7 +813,7 @@ def _hierarchical_neighbor_allreduce_nonblocking(
     dynamic_neighbors_enabled = True
     handle = getattr(mpi_lib, function)(tensor_buffer, output, self_weight, neighbor_weights,
                                         send_neighbors, dynamic_neighbors_enabled, enable_topo_check,
-                                        avg_computation, is_hierarchical,
+                                        weighted_average_computation, is_hierarchical,
                                         name.encode() if name is not None else "")
     _handle_map[handle] = (tensor_buffer, output)
     return handle
@@ -746,11 +824,11 @@ def _pair_gossip_nonblocking_function_factory(tensor):
 
 
 def _pair_gossip_nonblocking(tensor, output, target_rank, self_weight, pair_weight,
-                             avg_computation, name):
+                             weighted_average_computation, name):
     function = _check_function(
         _pair_gossip_nonblocking_function_factory, tensor)
     handle = getattr(mpi_lib, function)(tensor, output, target_rank,
-                                        self_weight, pair_weight, avg_computation,
+                                        self_weight, pair_weight, weighted_average_computation,
                                         name.encode() if name is not None else "")
     _handle_map[handle] = (tensor, output)
     return handle
@@ -810,14 +888,14 @@ def pair_gossip_nonblocking(tensor: torch.Tensor, target_rank: int, self_weight:
     if pair_weight is None and self_weight is None:
         pair_weight = 0
         self_weight = 0
-        avg_computation = True
+        weighted_average_computation = True
     elif pair_weight is not None and self_weight is not None:
-        avg_computation = False
+        weighted_average_computation = False
     else:
         raise ValueError("self_weight and pair_weight have to be set at same time.")
     output = tensor.new(tensor.shape)
     return _pair_gossip_nonblocking(tensor, output, target_rank, self_weight,
-                                    pair_weight, avg_computation, name)
+                                    pair_weight, weighted_average_computation, name)
 
 
 def poll(handle: int) -> bool:
@@ -1007,24 +1085,24 @@ def win_update(name: str,
         if not set(neighbor_weights.keys()).issubset(set(in_neighbor_ranks())):
             raise ValueError("The key of weights should only contain the ranks that belong to "
                              " in-neighbors and self rank.")
-        avg_computation = True
+        weighted_average_computation = True
 
     elif neighbor_weights is None and self_weight is None:
         if is_topo_weighted():
             topology = load_topology()
             self_weight, neighbor_weights = GetRecvWeights(topology, rank())
-            avg_computation = True
+            weighted_average_computation = True
         else:
             weight = 1.0/(len(in_neighbor_ranks())+1)
             self_weight = weight
             neighbor_weights = {r: weight for r in in_neighbor_ranks()}
-            avg_computation = False
+            weighted_average_computation = False
     else:
         raise ValueError("Arguments self_weight and neighbor_weights have to be presented at "
                          "the same time")
 
     if not getattr(mpi_lib, function)(tensor, name, self_weight, neighbor_weights,
-                                      reset, avg_computation, require_mutex):
+                                      reset, weighted_average_computation, require_mutex):
         raise RuntimeError("Cannot apply win_update on " + name)
     return tensor
 
@@ -1254,6 +1332,10 @@ def win_wait(handle: int) -> bool:
     return True
 
 
+def get_current_created_window_names() -> List[str]:
+    """Return the names of current created windows."""
+    return sorted(list(_win_map.keys()))
+
 def get_win_version(name: str) -> Dict[int, int]:
     """ Get the version of tensor stored in the win buffer.
 
